
## Copyright (c) 2001-2010, Scott D. Peckham
## July, August 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#-----------------------------------------------------------------------
#  NOTES:  This file defines a "Darcy layers" groundwater component
#          and related functions.  It inherits from the groundwater
#          "base class" in "GW_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class GW_darcy_layers
#
#      get_attribute()        # (10/26/11)
#      update()    ## (move all of the "update_*" methods to here??)
#
#  Functions:   (commented out)
#      Total_Darcy_Layer_Flow_VK()
#      Total_Subsurface_Flow()
#      Darcy_Layer_Seep_Rate()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import os

import cfg_files as cfg
import GW_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    g = GW_component()
    g.CCA   = False
    g.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory = tf_utils.TF_Test_Directory()
    cfg_prefix    = tf_utils.TF_Test_Case_Prefix()
    g.site_prefix = 'Treynor'  #############  
  
    g.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix )
    
##    g.initialize( cfg_prefix=cfg_prefix, mode="driver" )
    
    print 'nx =', g.nx
    print 'ny =', g.ny
    
#   unit_test()
#-----------------------------------------------------------------------
class GW_component(GW_base.GW_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'SatZoneDarcyLayers',
               'version':            '3.1',
               'model_name':         'Saturated_Zone_Darcy_Layers',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Satzone_Darcy_Layers.cfg.in',
               'cfg_extension':      '_satzone_darcy_layers.cfg',
               'cmt_var_prefix':     '/SatZoneDarcyLayers/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Satzone_Darcy_Layers.xml',
               'dialog_title':       'Saturated Zone: Darcy Layers Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------

#-----------------------------------------------------------------------
##def Total_Darcy_Layer_Flow_VK(gv, h, y, dw, ds, pIDs):
##
##    #-----------------------------------------------------
##    #NOTES:  gv = gw_vars = structure
##    #        z  = elevation of land surface [m]
##    #        h  = elevation of water table [m]
##    #        (z-h) = depth to water table [m]
##    #        Sh = water table slope [unitless]
##    #        K   = hydraulic conductivity [m/s];
##    #              each layer can have its own K grid,
##    #              represented as VK1, VK2, etc.
##    #        dw = element width [m]
##    #        ds = hor. Dist. between pixel and parent [m]
##    #        y  = wetted flow depth in each layer [m]
##    #             (could be a recycled local variable)
##    #        Q  = total Darcy-flow discharge [m^3/s]
##
##    #             (summed over all layers)
##    #        diff = (partial sum of soil thicknesses -
##    #                depth to water table)
##    #-----------------------------------------------------
##    
##    # FORWARD_FUNCTION Free_Surface_Slope
##    
##    #---------------------------------
##    #Compute water table slope from h
##    #---------------------------------
##    #NB!  h is assumed to be a grid.
##    #---------------------------------
##    #NB!  Q is zero where Sh is zero.
##    #-----------------------------------------
##    #NB!  Flow direction is still assumed to
##    #     be given by the DEM's D8 flow grid.
##    #-----------------------------------------
##    Sh = Free_Surface_Slope(float32(0.0), h, ds, pIDs)
##    
##    #----------------------------------------
##    #Compute wetted-depth, y, for each layer
##    #Now passed by caller.
##    #----------------------------------------
##    #** diff = -(z - h)
##    
##    #---------------------------------
##    #NB!  h is assumed to be a grid.
##    #---------------------------------
##    dims = idl_func.size(h, dimensions=True)
##    ncols = dims[0]
##    nrows = dims[1]
##    Q = zeros([nrows, ncols], dtype='Float32')
##    
##    #------------------------------------
##    #Add Q for each layer, via Darcy law
##    #------------------------------------
##    Q += (gv.VK1  * Sh * dw * y[0,:,:])
##    Q += (gv.VK2  * Sh * dw * y[1,:,:])
##    Q += (gv.VK3  * Sh * dw * y[2,:,:])
##    Q += (gv.VK4  * Sh * dw * y[3,:,:])
##    Q += (gv.VK5  * Sh * dw * y[4,:,:])
##    Q += (gv.VK6  * Sh * dw * y[5,:,:])
##    Q += (gv.VK7  * Sh * dw * y[6,:,:])
##    Q += (gv.VK8  * Sh * dw * y[7,:,:])
##    Q += (gv.VK9  * Sh * dw * y[8,:,:])
##    Q += (gv.VK10 * Sh * dw * y[9,:,:])
##    
##    return Q
##
###   Total_Darcy_Layer_Flow_VK
###-----------------------------------------------------------------------
##def Total_Subsurface_Flow(gv, h, y, dw, ds, pIDs):
##
##    #-------------------------------------------------------
##    #NOTES:  gv = gw_vars = structure
##    #        h  = elevation of water table [m]
##
##    #        Updates to y are also returned.
##    #-------------------------------------------------------
##    I2PY_expr = gv.method
##    if I2PY_expr == 0:    
##        Q_gw = float32(0.0)
##    elif I2PY_expr == 1:    
##        Q_gw = Total_Darcy_Layer_Flow(gv, h, y, dw, ds, pIDs)
##    elif I2PY_expr == 2:    
##        Q_gw = Total_Darcy_Layer_Flow_VK(gv, h, y, dw, ds, pIDs)
##        
##    else:
##        raise RuntimeError('no match found for expression')
##    
##    return Q_gw
##
##   Total_Subsurface_Flow
##-----------------------------------------------------------------------
##def Darcy_Layer_Seep_Rate(gv, h, z, y, Rg, dw, ds, da, pIDs, \
##                          p1, p2, p3, p4, p5, p6, p7, p8, \
##                          w1, w2, w3, w4, w5, w6, w7, w8):
##
##    #-------------------------------------------------
##    #Notes:  gv = gw_vars = structure
##    #        Bug fix on 7/19/05, gw_vars vs. gv used.
##    #
##    #7/19/05:  This function may no longer be in use
##    #          anywhere.  The call in the Seepage
##    #          function is commented out in favor of
##    #          a call to Total_Darcy_Layer_Flow.
##    #-------------------------------------------------
##
##    #------------------------------
##    #Get the vertical contribution
##    #See call to Precipitation.
##    #------------------------------
##    #Rg = 0.0
##
##    #-----------------------------
##    #Sum discharges of all layers
##    #-----------------------------
##    n_params = 25
##    I2PY_expr = gw_vars.method
##    if I2PY_expr == 0:    
##        Q_gw = float32(0.0)
##    elif I2PY_expr == 1:    
##        Q_gw = Total_Darcy_Layer_Flow(gv, h, y, dw, ds, pIDs)
##        
##    elif I2PY_expr == 2:    
##        Q_gw = Total_Darcy_Layer_Flow_VK(gv, h, y, dw, ds, pIDs)
##    else:
##        raise RuntimeError('no match found for expression')
##    
##    #--------------------------
##    #Print min and max of Q_gw
##    #--------------------------
##    Q_min = Q_gw.min()
##    Q_max = Q_gw.max()
##    TF_Print('   Q_min = ' + str(Q_min))
##    TF_Print('   Q_max = ' + str(Q_max))
##    
##    #--------------------------------
##    #Overwrite h & y with new values
##    #Need to pass gw_vars.
##    #--------------------------------
##    gv = Update_Water_Table(h, y, Q_gw, Rg, da, gv, \
##                            p1, p2, p3, p4, p5, p6, p7, p8, \
##                            w1, w2, w3, w4, w5, w6, w7, w8)
##    
##    #--------------------------------------------
##    #y is now updated in previous routine, but
##    #is initialized with Wetted_Thicknesses fcn.
##    #--------------------------------------------
##    #*** y = Wetted_Thicknesses(gv, z, h)
##    
##    #--------------
##    #For debugging
##    #--------------
##    #h_min = min(h, max=h_max)
##    
##    #print,'h_min = ',h_min
##    #print,'h_max = ',h_max
##    #------------------------------
##    #z_min = min(z, max=z_max)
##    #print,'z_min = ',z_min
##    #print,'z_max = ',z_max
##    
##    #------------------------
##    #Compute the "seep rate"
##    #Can be pos. or neg. ?
##    #------------------------
##    dh_dt = (h - h_last) / gv.dt
##    GW = (h > z) * dh_dt
##    #*** GW = (h gt z) * (dh_dt > 0.0)
##    
##    #------------------------
##    #Print min and max of GW
##    #------------------------
##    gw_min = gw.min()
##    gw_max = gw.max()
##    TF_Print('   GW_min = ' + str(gw_min))
##    TF_Print('   GW_max = ' + str(gw_max))
##    
##    return GW
##
##   Darcy_Layer_Seep_Rate
##-----------------------------------------------------------------------
##
##
