
## Copyright (c) 2001-2010, Scott D. Peckham
## January, May, July 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#-----------------------------------------------------------------------
#  NOTES:  This file defines a "kinematic wave" channel flow component
#          and related functions.  It inherits from the channels
#          "base class" in "channels_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class channels_component
#
#      get_attribute()  # (10/26/11)
#      update_velocity()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import cfg_files as cfg
import channels_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    c = channels_component()
    c.CCA   = False
    c.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
##    cfg_directory = tf_utils.TF_Test_Directory()

    ## cfg_directory = '/data/progs/topoflow/3.0/data/test_plane_canal/'
    cfg_directory = '/Applications/TopoFlow/Data/Test_Plane_Canal/'
    cfg_prefix    = 'Case5'
    c.site_prefix = 'plane' ########
    
    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize( mode="driver" )
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update( time_sec )
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix, n_steps=100 )

##    P = c.get_port_data('P', c.mp)
##    print ' '
##    print 'Reading from "met" port:'
##    print 'P =', P
##    print ' '
    
##    print 'KINEMATIC WAVE =', c.KINEMATIC_WAVE
##    print 'DIFFUSIVE WAVE =', c.DIFFUSIVE_WAVE
##    print 'DYNAMIC WAVE   =', c.DYNAMIC_WAVE
##    print ' '
##    print 'nx =', c.nx
##    print 'ny =', c.ny
    
#   unit_test()
#-----------------------------------------------------------------------
class channels_component(channels_base.channels_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        #-----------------------------------------------------------
        # This is done in channels_base.set_computed_input_vars()
        #-----------------------------------------------------------
        # self.KINEMATIC_WAVE = True
        # self.DIFFUSIVE_WAVE = False
        # self.DYNAMIC_WAVE   = False

        #----------------
        # New. 10/26/11
        #----------------
        map = {'comp_name':          'ChannelsKinWave',
               'version':            '3.1',
               'model_name':         'Channels_Kinematic_Wave',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Channels_Kinematic_Wave.cfg.in',
               'cfg_extension':      '_channels_kinematic_wave.cfg',
               'cmt_var_prefix':     '/ChannelsKinWave/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Channels_Kinematic_Wave.xml',
               'dialog_title':       'Channels: Kinematic Wave Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'uniform',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def update_velocity(self):

        #---------------------------------------------------------
        # Notes: Compute u from d and S_bed.  (7/13/05 version)

        #        nval   = Manning's n values (grid)
        #        z0_val = z0 roughness values (grid)
        #        width  = channel bottom widths (grid)
        #        angle  = channel bank angles (grid)

        #        Could use slopes in cp also, but S_bed has
        #        been modified from those values to impose a
        #        minimum slope that is nonzero.

        #        Rh = hydraulic radius (trapezoid here)

        #        S = S_bed  for KINEMATIC_WAVE option.
        #        S = S_free for DIFFUSIVE_WAVE_option.
        #---------------------------------------------------------
            
        #------------------------
        # Use Manning's formula
        #------------------------
        if (self.MANNING):    
            self.u = self.manning_formula(self.S_bed)
        
        #--------------------------------------
        # Use the Logarithmic Law of the Wall
        #--------------------------------------
        if (self.LAW_OF_WALL):    
            self.u = self.law_of_the_wall(self.S_bed)

        # print '(umin, umax) =', self.u.min(), self.u.max()
        
    #   update_velocity()                       
    #-------------------------------------------------------------------


         
