
## Copyright (c) 2010, Scott D. Peckham
## Jan 2010   (started from diversions_base.py)

#---------------------------------------------------------------------
# Notes:  cp.update_discharge() now calls dp.update(). (2/1/10)
#
#---------------------------------------------------------------------
#
#  unit_test()
#
#  class diversions_component:  (inherits from diversions_base.py)
#      update()
#----------------------------
#      read_input_files()
#      read_source_data()
#      read_sink_data()
#      read_canal_data()
#----------------------------
#      update_sources()
#      update_sinks()
#      update_canals()
#
#---------------------------------------------------------------------

from numpy import *
import numpy

import glob

import cfg_files as cfg
import diversions_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    #------------------------------------------
    # These functions adjust for the platform
    # and can be changed in "tf_utils.py".
    #------------------------------------------
    directory   = tf_utils.TF_Test_Directory()
    data_prefix = tf_utils.TF_Test_Data_Prefix()
    case_prefix = tf_utils.TF_Test_Case_Prefix()
    
    d = diversions_component()
    d.CCA = False
    
    d.initialize(directory=directory,
                 data_prefix=data_prefix,
                 case_prefix=case_prefix, mode="main")

    print 'use_sources =', d.use_sources
    print 'use_sinks   =', d.use_sinks
    print 'use_canals  =', d.use_canals
    print ' '
    print 'source_file =', d.source_file
    print 'sink_file   =', d.sink_file
    print 'canal_file  =', d.canal_file
    
    print 'Finished with unit_test().'

#   unit_test()
#---------------------------------------------------------------------
class diversions_component(diversions_base.diversions_component):
    
    #--------------------------------------------------------------------------
    def update(self, time_seconds=None):

        #-----------------------------------------------------------
        # Notes:  This new method is written so that the Channels
        #         component only retrieves a small amount of data
        #         from the Diversions component.  The old method
        #         required the Diversion component to retrieve the
        #         entire Q or vol grid from Channel component at
        #         each timestep, alter it, and then pass it back.
        #        (02/12/10)
        #-----------------------------------------------------------   
        if (self.method == 0): return
        self.status = 'updating'  # (OpenMI 2.0 convention)
        
        #-----------------------------------------------
        # Update self.vol with inputs/outputs from all
        # sources, sinks and diversions
        #-----------------------------------------------
        self.update_sources()
        self.update_sinks()
        self.update_canals()

        #------------------------
        # Update internal clock
        #------------------------
        self.update_time()        
        self.status = 'updated'  # (OpenMI 2.0 convention)
        
    #   update()
    #--------------------------------------------------------------------------                
    def read_source_data(self):

        #------------------------------------------------------------
        # Notes:  Assume that source_file contains key-value pairs,
        #         starting with "n_sources:", "nt_max" and "dt:",
        #         followed by "n_sources" blocks of the form:
        #
        #         source_ID: (source pixel ID as long integer)
        #         nt:        (number of discharge (Q) values)
        #         Q:         (vector of discharges in m^3/s)
        #------------------------------------------------------------
        if not((self.method == 1) and self.use_sources):  return
        
        #-----------------------------
        # Can source_file be found ?
        #-----------------------------
        FOUND = tf_utils.file_exists( self.source_file )
        if not(FOUND):
            self.use_sources = False
            return

        #-------------------------
        # Open the "source_file"
        #-------------------------
        file_unit = open(self.source_file, 'r')

        #--------------------------------------------------
        # Read number of sources, max number of timesteps
        # for any source and the common timestep, dt
        #--------------------------------------------------
        n_sources = cfg.read_value(file_unit, dtype='Int32')
        nt_max    = cfg.read_value(file_unit, dtype='Int32')
        dt        = cfg.read_value(file_unit, dtype='Float64')

        #--------------------
        # Initialize arrays
        #--------------------
        self.source_IDs    = zeros([n_sources], dtype='Int32')
        self.nt_sources    = zeros([n_sources], dtype='Int32')
        self.Q_sources_all = zeros([n_sources, nt_max], dtype='Float64')
        self.n_sources     = n_sources

        self.nt_max_sources = nt_max
        
        #-----------------------------------
        # Read information for each source
        #-----------------------------------
        for k in xrange(n_sources):
            source_ID = cfg.read_value(file_unit, dtype='Int32')
            nt        = cfg.read_value(file_unit, dtype='Int32')
            Q_values  = cfg.read_list_after_key(file_unit, dtype='Float64')
            #---------------------------------------------------------------
            nQ        = size(Q_values)
            print 'Diversions component: Read', nQ, 'Q_values for source.'
            #---------------------------------------------------------------            
            self.source_IDs[k]     = source_ID
            self.nt_sources[k]     = nt
            self.Q_sources_all[k,0:nt] = Q_values
 
        #-----------------------
        # Close the input file
        #-----------------------
        file_unit.close()
        
        #-----------------------------------------
        # Save initial discharge for each source
        #--------------------------------------------------------
        # Assumes dp.update() will be called after cp.update().
        #--------------------------------------------------------
        self.Q_sources = self.Q_sources_all[:, 0]
        # print 'type(Q_sources) =', type(Q_sources), 'in read_source_data().'
        
        #------------------------------------------------------
        # Copy information into the Channels component
        # This may be handled differently later on. (2/16/10)
        #------------------------------------------------------
        print 'Diversions component: Setting source info in CHANNELS...'
        self.cp.set_scalar_long('use_sources', self.use_sources)
        self.cp.set_vector_long('source_IDs',  self.source_IDs)
        
    #   read_source_data()
    #--------------------------------------------------------------------------
    def read_sink_data(self):
        
        #------------------------------------------------------------
        # Notes:  Assume that source_file contains key-value pairs,
        #         starting with "n_sinks:", "nt_max" and "dt:",
        #         followed by "n_sinks" blocks of the form:
        #
        #         sink_ID: (sink pixel ID as long integer)
        #         nt:      (number of discharge (Q) values)
        #         Q:       (vector of discharges in m^3/s)
        #------------------------------------------------------------
        if not((self.method == 1) and self.use_sinks):  return
    
        #---------------------------
        # Can sink_file be found ?
        #---------------------------
        FOUND = tf_utils.file_exists( self.sink_file )
        if not(FOUND):
            self.use_sinks = False
            return

        #-----------------------
        # Open the "sink_file"
        #-----------------------
        file_unit = open(self.sink_file, 'r')

        #------------------------------------------------
        # Read number of sinks, max number of timesteps
        # for any sink and the common timestep, dt
        #------------------------------------------------
        n_sinks = cfg.read_value(file_unit, dtype='Int32')
        nt_max  = cfg.read_value(file_unit, dtype='Int32')
        dt      = cfg.read_value(file_unit, dtype='Float64')

        #--------------------
        # Initialize arrays
        #--------------------
        self.sink_IDs    = zeros([n_sinks], dtype='Int32')
        self.nt_sinks    = zeros([n_sinks], dtype='Int32')
        self.Q_sinks_all = zeros([n_sinks, nt_max], dtype='Float64')
        self.n_sinks     = n_sinks
        
        self.nt_max_sinks = nt_max
        
        #---------------------------------
        # Read information for each sink
        #---------------------------------
        for k in xrange(n_sinks):
            sink_ID  = cfg.read_value(file_unit, dtype='Int32')
            nt       = cfg.read_value(file_unit, dtype='Int32')
            Q_values = cfg.read_list_after_key(file_unit, dtype='Float64')
            #---------------------------------------------------------------
            nQ        = size(Q_values)
            print 'Diversions component: Read', nQ, 'Q_values for sink.'
            #--------------------------------------------------------------- 
            self.sink_IDs[k]     = sink_ID
            self.nt_sinks[k]     = nt
            self.Q_sinks_all[k,0:nt] = Q_values

        #-----------------------
        # Close the input file
        #-----------------------
        file_unit.close()              

        #---------------------------------------
        # Save initial discharge for each sink
        #--------------------------------------------------------
        # Assumes dp.update() will be called after cp.update().
        #--------------------------------------------------------
        self.Q_sinks = self.Q_sinks_all[:, 0]

        #------------------------------------------------------
        # Copy information into the Channels component
        # This may be handled differently later on. (2/16/10)
        #------------------------------------------------------
        print 'Diversions component: Setting sink info in CHANNELS...'
        self.cp.set_scalar_long('use_sinks', self.use_sinks)
        self.cp.set_vector_long('sink_IDs',  self.sink_IDs)

    #   read_sink_data()
    #--------------------------------------------------------------------------
    def read_canal_data(self):

        #-------------------------------------------------------------------
        # Notes:  Assume that canal_file contains key-value pairs,
        #         starting with "n_canals:" and followed by "n_canals"
        #         blocks of the form:
        #            canal_in_ID:  (pixel ID as long integer)
        #            canal_out_ID: (pixel ID as long integer)
        #            Q_fraction:   (fraction to take from in_ID in [0,1])
        #            travel_time:  (canal travel time, in minutes)
        #
        #         nt_canals is computed as ceil(travel_time / cp.dt)
        #-------------------------------------------------------------------
        # Note:  Q_canals is same at upstream and downstream ends, but the
        #        downstream end lags the upstream end by the travel time
        #        from in_ID to out_ID.  As a result, the duration and Q
        #        vector for the downstream end are computed from those of
        #        the upstream end, and the travel time, td, as:
        #            Q_out   = [0,  Q_in]
        #            dur_out = [td, dur_in]
        #            dur_sum_out = [0, dur_sum_in] + td
        #
        #        Rather than create the dur_sum_canals_out and
        #        Q_canals_out vectors, can construct them in Update_Canals.
        #-------------------------------------------------------------------
        if not((self.method == 1) and self.use_canals):  return
    
        #---------------------------
        # Can canal_file be found ?
        #---------------------------
        FOUND = tf_utils.file_exists( self.canal_file )
        if not(FOUND):
            self.use_canals = False
            return

        #------------------------
        # Open the "canal_file"
        #------------------------
        file_unit = open(self.canal_file, 'r')

        #------------------------
        # Read number of canals
        #------------------------
        n_canals      = cfg.read_value(file_unit, dtype='Int32')
        self.n_canals = n_canals

        #--------------------
        # Initialize arrays
        #--------------------
        self.canal_in_IDs      = zeros([n_canals], dtype='Int32')
        self.canal_out_IDs     = zeros([n_canals], dtype='Int32')
        self.canal_Q_fractions = zeros([n_canals], dtype='Float64')
        self.canal_times       = zeros([n_canals], dtype='Float64')
        
        #----------------------------------
        # Read information for each canal
        #----------------------------------
        for k in xrange(n_canals):
            canal_in_ID  = cfg.read_value(file_unit, dtype='Int32')
            canal_out_ID = cfg.read_value(file_unit, dtype='Int32')
            Q_fraction   = cfg.read_value(file_unit, dtype='Float64')
            travel_time  = cfg.read_value(file_unit, dtype='Float64')
            #----------------------------------------------------------
            self.canal_in_IDs[k]      = canal_in_ID
            self.canal_out_IDs[k]     = canal_out_ID
            self.canal_Q_fractions[k] = Q_fraction
            self.canal_times[k]       = travel_time

        #--------------------------------------------------------
        # Compute "nt_canals", which is the number of timesteps
        # it takes for flow to travel from end to end.
        #--------------------------------------------------------
        # This depends on "self.dt", which is set to self.cp.dt
        # by self.initialize().  It can vary between runs.
        #--------------------------------------------------------
        self.nt_canals = numpy.ceil(self.canal_times / self.dt)
        
        #-----------------------
        # Close the input file
        #-----------------------
        file_unit.close()

        #-----------------------------------------------------
        # Create a 2D array to store the discharge values as
        # they are moving toward downstream end of canal.
        #-----------------------------------------------------
        # update_canals() will "roll" this array downstream
        # by one array element each time step
        #-----------------------------------------------------
        nt_max       = int(self.nt_canals.max())
        nt_min       = int(self.nt_canals.min())
        self.canal_Q = numpy.zeros([n_canals, nt_max], dtype='Float64')
        self.nt_max  = nt_max
        print 'Diversions component: Min steps per canal =', nt_min
        print 'Diversions component: Max steps per canal =', nt_max
        
        #-----------------------------------------------
        # Save initial discharge *leaving* each canal.
        # Channels component will set Q_canals_in.
        #--------------------------------------------------------
        # Assumes dp.update() will be called after cp.update().
        #--------------------------------------------------------
        self.Q_canals_out = numpy.zeros(n_canals)

        #------------------------------------------------------
        # Copy information into the Channels component
        # This may be handled differently later on. (2/16/10)
        #------------------------------------------------------
        print 'Diversions component: Setting canal info in CHANNELS...'
        self.cp.set_scalar_long('use_canals',          self.use_canals)
        self.cp.set_vector_long('canal_in_IDs',        self.canal_in_IDs)
        self.cp.set_vector_long('canal_out_IDs',       self.canal_out_IDs)
        self.cp.set_vector_double('canal_Q_fractions', self.canal_Q_fractions)
        
    #   read_canal_data()
    #--------------------------------------------------------------------------
    def update_sources(self):

        #---------------------------------------------------------
        # Notes: This function avoids loops in favor of array
        #        operations to increase speed.
        #---------------------------------------------------------
        #        The number of Q-values for each source ID are
        #        stored as "self.nt_sources".  However, for any
        #        given source ID, the Q-values beyond that index
        #        are set to zero.
        #---------------------------------------------------------        
        if not(self.use_sources): return
        
        #-------------------------------------------  
        # Update discharge, Q, for every source ID
        #-------------------------------------------
        if (self.time_index < self.nt_max_sources):
            self.Q_sources = self.Q_sources_all[ :, self.time_index ]
        else:
            self.Q_sources = zeros(self.n_sources)

        #--------------
        # For testing
        #--------------
##        print 'self.Q_sources ='
##        print self.Q_sources
        
    #   update_sources()
    #--------------------------------------------------------------------------
    def update_sinks(self):

        #-------------------------------------------------------
        # Notes: This function avoids loops in favor of array
        #        operations to increase speed.
        #-------------------------------------------------------
        #        The number of Q-values for each sink ID are
        #        stored as "self.nt_sinks".  However, for any
        #        given sink ID, the Q-values beyond that index
        #        are set to zero.
        #-------------------------------------------------------  
        if not(self.use_sinks): return

        #-----------------------------------------  
        # Update discharge, Q, for every sink ID
        #-----------------------------------------
        # Make sure sink cannot produce negative
        # discharge values.
        #-----------------------------------------
        if (self.time_index < self.nt_max_sinks):
            self.Q_sinks = self.Q_sinks_all[ :, self.time_index ]
        else:
            self.Q_sinks = zeros(self.n_sinks)
                
    #   update_sinks()
    #--------------------------------------------------------------------------
    def update_canals(self):

        #----------------------------------------------------------
        # Notes: Before 2/1/10, TopoFlow would update the channel
        #        component before the diversion component.  Now
        #        cp.update_discharge() calls dp.update() itself.
        #
        #        (2/16/10) Tested for single canal and seems to
        #        work as intended.
        #
        # NB!    Flow volumes are incremented (by the function
        #        update_flow_volume(), but discharges are
        #        always recomputed from d, v and channel geom.
        #        So changes to cp.Q by calling dp.update()
        #        after cp.update() would be lost.
        #
        #        cp.update() computes variables in this order:
        #           update_R()
        #           update_discharge()    (using d and v)
        #           update_flow_volume()  (using Q and R)
        #           update_flow_depth()   (using vol)
        #           update_velocity()
        #----------------------------------------------------------
        # Notes: This function avoids loops in favor of array
        #        operations to increase speed.
        #----------------------------------------------------------
        # Notes: In this version, the Channels component uses
        #        canal_Q_fractions, canal_in_IDs and its own
        #        Q grid to compute Q_canals_in.  It then sets
        #        this into the Diversion component.
        #----------------------------------------------------------        
        if not(self.use_canals): return
        
        #----------------------------------------------
        # Add specified discharge (Q_in) to upstream
        # end of each canal (at array index 0)
        #----------------------------------------------
        # NB! Q_canals_in will be set by the Channels
        #     component.
        #----------------------------------------------
        self.canal_Q[:, 0] = self.Q_canals_in   ########
        
        #------------------------------------------------
        # Get "Q" at downstream end of each canal.
        # It will be zero until flow has had time to
        # travel the distance.
        #------------------------------------------------
        # NB!  Each canal can have a different travel
        # time and therefore its own "nt_canal" value.
        #------------------------------------------------
        # NB! Q_canals_out will be retrieved by the
        #     Channels component.
        #------------------------------------------------
        nc  = self.n_canals
        self.Q_canals_out = numpy.empty(nc, dtype='Float32')
        for k in xrange(nc):
            nt_k  = self.nt_canals[k]
            self.Q_canals_out[k] = self.canal_Q[:, nt_k-1]
        ## self.canal_Q[:, nt_k:] = 0.0  # (not necessary)

        #--------------
        # For testing
        #--------------
##        print 'self.canal_Q ='
##        print self.canal_Q
##        print 'self.Q_canals_out ='
##        print self.Q_canals_out
##        print ' '
        
        #----------------------------------------------------
        # "Roll" the canal_Q array downstream (along its
        # 2nd index, with index = 1) by one array element
        # in each time step. "canal_Q" starts with zeros;
        # i.e.  canal_Q = zeros([n_canals, n_steps])
        #----------------------------------------------------
        # In next call to update_canals(), we'll replace
        # the first Q-value in each canal.
        #----------------------------------------------------
        self.canal_Q = numpy.roll( self.canal_Q, 1, axis=1 )
    
    #   update_canals()
    #--------------------------------------------------------------------------


    

